@@ -1,5 +1,5 @@
 /*
- *  Copyright (c) 1999-2002 Sendmail, Inc. and its suppliers.
+ *  Copyright (c) 1999-2003 Sendmail, Inc. and its suppliers.
  *	All rights reserved.
  *
  * By using this file, you agree to the terms and conditions set
@@ -9,7 +9,7 @@
  */
 
 #include <sm/gen.h>
-SM_RCSID("@(#)$Id: engine.c,v 8.109.2.5 2003/08/04 18:14:33 ca Exp $")
+SM_RCSID("@(#)$Id: engine.c,v 8.119 2003/12/02 18:53:57 ca Exp $")
 
 #include "libmilter.h"
 
@@ -59,13 +59,14 @@ typedef struct cmdfct_t cmdfct;
 #define	CT_END		0x0008	/* start replying */
 
 /* index in macro array: macros only for these commands */
-#define	CI_NONE	(-1)
-#define	CI_CONN	0
-#define	CI_HELO	1
-#define	CI_MAIL	2
-#define	CI_RCPT	3
-#if CI_RCPT >= MAX_MACROS_ENTRIES
-ERROR: do not compile with CI_RCPT >= MAX_MACROS_ENTRIES
+#define	CI_NONE		(-1)
+#define	CI_CONN		0
+#define	CI_HELO		1
+#define	CI_MAIL		2
+#define CI_RCPT		3
+#define CI_EOM		4
+#if CI_EOM >= MAX_MACROS_ENTRIES
+ERROR: do not compile with CI_EOM >= MAX_MACROS_ENTRIES
 #endif
 
 /* function prototypes */
@@ -79,6 +80,12 @@ static int	st_helo __P((genarg *));
 static int	st_header __P((genarg *));
 static int	st_sender __P((genarg *));
 static int	st_rcpt __P((genarg *));
+#if SMFI_VERSION > 2
+static int	st_unknown __P((genarg *));
+#endif /* SMFI_VERSION > 2 */
+#if SMFI_VERSION > 3
+static int	st_data __P((genarg *));
+#endif /* SMFI_VERSION > 3 */
 static int	st_eoh __P((genarg *));
 static int	st_quit __P((genarg *));
 static int	sendreply __P((sfsistat, socket_t, struct timeval *, SMFICTX_PTR));
@@ -95,13 +102,15 @@ static int	dec_arg2 __P((char *, size_t, char **, char **));
 #define ST_HELO	3	/* helo */
 #define ST_MAIL	4	/* mail from */
 #define ST_RCPT	5	/* rcpt to */
-#define ST_HDRS	6	/* headers */
-#define ST_EOHS	7	/* end of headers */
-#define ST_BODY	8	/* body */
-#define ST_ENDM	9	/* end of message */
-#define ST_QUIT	10	/* quit */
-#define ST_ABRT	11	/* abort */
-#define ST_LAST	ST_ABRT
+#define ST_DATA	6	/* data */
+#define ST_HDRS	7	/* headers */
+#define ST_EOHS	8	/* end of headers */
+#define ST_BODY	9	/* body */
+#define ST_ENDM	10	/* end of message */
+#define ST_QUIT	11	/* quit */
+#define ST_ABRT	12	/* abort */
+#define ST_UNKN 13	/* unknown SMTP command */
+#define ST_LAST	ST_UNKN	/* last valid state */
 #define ST_SKIP	15	/* not a state but required for the state table */
 
 /* in a mail transaction? must be before eom according to spec. */
@@ -118,19 +127,25 @@ static int	dec_arg2 __P((char *, size_t, char **, char **));
 
 #define MI_MASK(x)	(0x0001 << (x))	/* generate a bit "mask" for a state */
 #define NX_INIT	(MI_MASK(ST_OPTS))
-#define NX_OPTS	(MI_MASK(ST_CONN))
-#define NX_CONN	(MI_MASK(ST_HELO) | MI_MASK(ST_MAIL))
-#define NX_HELO	(MI_MASK(ST_HELO) | MI_MASK(ST_MAIL))
-#define NX_MAIL	(MI_MASK(ST_RCPT) | MI_MASK(ST_ABRT))
-#define NX_RCPT	(MI_MASK(ST_HDRS) | MI_MASK(ST_EOHS) | \
+#define NX_OPTS	(MI_MASK(ST_CONN) | MI_MASK(ST_UNKN))
+#define NX_CONN	(MI_MASK(ST_HELO) | MI_MASK(ST_MAIL) | MI_MASK(ST_UNKN))
+#define NX_HELO	(MI_MASK(ST_HELO) | MI_MASK(ST_MAIL) | MI_MASK(ST_UNKN))
+#define NX_MAIL	(MI_MASK(ST_RCPT) | MI_MASK(ST_ABRT) | MI_MASK(ST_UNKN))
+#define NX_RCPT	(MI_MASK(ST_HDRS) | MI_MASK(ST_EOHS) | MI_MASK(ST_DATA) | \
 		 MI_MASK(ST_BODY) | MI_MASK(ST_ENDM) | \
-		 MI_MASK(ST_RCPT) | MI_MASK(ST_ABRT))
+		 MI_MASK(ST_RCPT) | MI_MASK(ST_ABRT) | MI_MASK(ST_UNKN))
+#define NX_DATA	(MI_MASK(ST_EOHS) | MI_MASK(ST_HDRS) | MI_MASK(ST_ABRT))
 #define NX_HDRS	(MI_MASK(ST_EOHS) | MI_MASK(ST_HDRS) | MI_MASK(ST_ABRT))
 #define NX_EOHS	(MI_MASK(ST_BODY) | MI_MASK(ST_ENDM) | MI_MASK(ST_ABRT))
 #define NX_BODY	(MI_MASK(ST_ENDM) | MI_MASK(ST_BODY) | MI_MASK(ST_ABRT))
-#define NX_ENDM	(MI_MASK(ST_QUIT) | MI_MASK(ST_MAIL))
+#define NX_ENDM	(MI_MASK(ST_QUIT) | MI_MASK(ST_MAIL) | MI_MASK(ST_UNKN))
 #define NX_QUIT	0
 #define NX_ABRT	0
+#define NX_UNKN (MI_MASK(ST_HELO) | MI_MASK(ST_MAIL) | \
+		 MI_MASK(ST_RCPT) | MI_MASK(ST_ABRT) | \
+		 MI_MASK(ST_DATA) | \
+		 MI_MASK(ST_BODY) | MI_MASK(ST_UNKN) | \
+		 MI_MASK(ST_ABRT) | MI_MASK(ST_QUIT))
 #define NX_SKIP MI_MASK(ST_SKIP)
 
 static int next_states[] =
@@ -141,12 +156,14 @@ static int next_states[] =
 	NX_HELO,
 	NX_MAIL,
 	NX_RCPT,
+	NX_DATA,
 	NX_HDRS,
 	NX_EOHS,
 	NX_BODY,
 	NX_ENDM,
 	NX_QUIT,
-	NX_ABRT
+	NX_ABRT,
+	NX_UNKN
 };
 
 /* commands received by milter */
@@ -156,14 +173,20 @@ static cmdfct cmds[] =
 {SMFIC_MACRO,	CM_ARGV, ST_NONE,  CT_KEEP,	CI_NONE, st_macros	},
 {SMFIC_BODY,	CM_ARG1, ST_BODY,  CT_CONT,	CI_NONE, st_bodychunk	},
 {SMFIC_CONNECT,	CM_ARG2, ST_CONN,  CT_CONT,	CI_CONN, st_connectinfo	},
-{SMFIC_BODYEOB,	CM_ARG1, ST_ENDM,  CT_CONT,	CI_NONE, st_bodyend	},
+{SMFIC_BODYEOB,	CM_ARG1, ST_ENDM,  CT_CONT,	CI_EOM,  st_bodyend	},
 {SMFIC_HELO,	CM_ARG1, ST_HELO,  CT_CONT,	CI_HELO, st_helo	},
 {SMFIC_HEADER,	CM_ARG2, ST_HDRS,  CT_CONT,	CI_NONE, st_header	},
 {SMFIC_MAIL,	CM_ARGV, ST_MAIL,  CT_CONT,	CI_MAIL, st_sender	},
 {SMFIC_OPTNEG,	CM_ARGO, ST_OPTS,  CT_CONT,	CI_NONE, st_optionneg	},
 {SMFIC_EOH,	CM_ARG0, ST_EOHS,  CT_CONT,	CI_NONE, st_eoh		},
 {SMFIC_QUIT,	CM_ARG0, ST_QUIT,  CT_END,	CI_NONE, st_quit	},
+#if SMFI_VERSION > 3
+{SMFIC_DATA,	CM_ARG0, ST_DATA,  CT_CONT,	CI_NONE, st_data	},
+#endif /* SMFI_VERSION > 3 */
 {SMFIC_RCPT,	CM_ARGV, ST_RCPT,  CT_IGNO,	CI_RCPT, st_rcpt	}
+#if SMFI_VERSION > 2
+,{SMFIC_UNKNOWN,CM_ARG1, ST_UNKN,  CT_IGNO,	CI_NONE, st_unknown	}
+#endif /* SMFI_VERSION > 2 */
 };
 
 /* additional (internal) reply codes */
@@ -599,7 +622,7 @@ st_connectinfo(g)
 	l = g->a_len;
 	while (s[i] != '\0' && i <= l)
 		++i;
-	if (i >= l)
+	if (i + 1 >= l)
 		return _SMFIS_ABORT;
 
 	/* Move past trailing \0 in host string */
@@ -608,9 +631,7 @@ st_connectinfo(g)
 	(void) memset(&sockaddr, '\0', sizeof sockaddr);
 	if (family != SMFIA_UNKNOWN)
 	{
-		(void) memcpy((void *) &port, (void *) (s + i),
-			      sizeof port);
-		if ((i += sizeof port) >= l)
+		if (i + sizeof port >= l)
 		{
 			smi_log(SMI_LOG_ERR,
 				"%s: connect[%d]: wrong len %d >= %d",
@@ -618,6 +639,9 @@ st_connectinfo(g)
 				(int) g->a_ctx->ctx_id, (int) i, (int) l);
 			return _SMFIS_ABORT;
 		}
+		(void) memcpy((void *) &port, (void *) (s + i),
+			      sizeof port);
+		i += sizeof port;
 
 		/* make sure string is terminated */
 		if (s[l - 1] != '\0')
@@ -686,6 +710,7 @@ st_connectinfo(g)
 	return (*fi_connect)(g->a_ctx, g->a_buf,
 			     family != SMFIA_UNKNOWN ? &sockaddr : NULL);
 }
+
 /*
 **  ST_EOH -- end of headers
 **
@@ -709,6 +734,33 @@ st_eoh(g)
 		return (*fi_eoh)(g->a_ctx);
 	return SMFIS_CONTINUE;
 }
+
+#if SMFI_VERSION > 3
+/*
+**  ST_DATA -- DATA command
+**
+**	Parameters:
+**		g -- generic argument structure
+**
+**	Returns:
+**		continue or filter-specified value
+*/
+
+static int
+st_data(g)
+	genarg *g;
+{
+	sfsistat (*fi_data) __P((SMFICTX *));
+
+	if (g == NULL)
+		return _SMFIS_ABORT;
+	if (g->a_ctx->ctx_smfi != NULL &&
+	    (fi_data = g->a_ctx->ctx_smfi->xxfi_data) != NULL)
+		return (*fi_data)(g->a_ctx);
+	return SMFIS_CONTINUE;
+}
+#endif /* SMFI_VERSION > 3 */
+
 /*
 **  ST_HELO -- helo/ehlo command
 **
@@ -729,7 +781,12 @@ st_helo(g)
 	mi_clr_macros(g->a_ctx, g->a_idx + 1);
 	if (g->a_ctx->ctx_smfi != NULL &&
 	    (fi_helo = g->a_ctx->ctx_smfi->xxfi_helo) != NULL)
+	{
+		/* paranoia: check for terminating '\0' */
+		if (g->a_len == 0 || g->a_buf[g->a_len - 1] != '\0')
+			return MI_FAILURE;
 		return (*fi_helo)(g->a_ctx, g->a_buf);
+	}
 	return SMFIS_CONTINUE;
 }
 /*
@@ -809,6 +866,34 @@ st_rcpt(g)
 {
 	ARGV_FCT(fi_envrcpt, xxfi_envrcpt, CI_RCPT)
 }
+
+#if SMFI_VERSION > 2
+/*
+**  ST_UNKNOWN -- unrecognized or unimplemented command
+**
+**	Parameters:
+**		g -- generic argument structure
+**
+**	Returns:
+**		continue or filter-specified value
+*/
+
+static int
+st_unknown(g)
+	genarg *g;
+{
+	sfsistat (*fi_unknown) __P((SMFICTX *, char *));
+
+	if (g == NULL)
+		return _SMFIS_ABORT;
+	mi_clr_macros(g->a_ctx, g->a_idx + 1);
+	if (g->a_ctx->ctx_smfi != NULL &&
+	    (fi_unknown = g->a_ctx->ctx_smfi->xxfi_unknown) != NULL)
+		return (*fi_unknown)(g->a_ctx, g->a_buf);
+	return SMFIS_CONTINUE;
+}
+#endif /* SMFI_VERSION > 2 */
+
 /*
 **  ST_MACROS -- deal with macros received from the MTA
 **
@@ -847,6 +932,9 @@ st_macros(g)
 	  case SMFIC_RCPT:
 		i = CI_RCPT;
 		break;
+	  case SMFIC_BODYEOB:
+		i = CI_EOM;
+		break;
 	  default:
 		free(argv);
 		return _SMFIS_FAIL;
@@ -1089,10 +1177,16 @@ dec_argv(buf, len)
 	for (i = 0, elem = 0; i < len && elem < nelem; i++)
 	{
 		if (buf[i] == '\0')
-			s[++elem] = &(buf[i + 1]);
+		{
+			++elem;
+			if (i + 1 >= len)
+				s[elem] = NULL;
+			else
+				s[elem] = &(buf[i + 1]);
+		}
 	}
 
-	/* overwrite last entry */
+	/* overwrite last entry (already done above, just paranoia) */
 	s[elem] = NULL;
 	return s;
 }
@@ -1117,6 +1211,9 @@ dec_arg2(buf, len, s1, s2)
 {
 	size_t i;
 
+	/* paranoia: check for terminating '\0' */
+	if (len == 0 || buf[len - 1] != '\0')
+		return MI_FAILURE;
 	*s1 = buf;
 	for (i = 1; i < len && buf[i] != '\0'; i++)
 		continue;
