@@ -1,5 +1,5 @@
 /*
- * Copyright (c) 1998-2002 Sendmail, Inc. and its suppliers.
+ * Copyright (c) 1998-2003 Sendmail, Inc. and its suppliers.
  *	All rights reserved.
  * Copyright (c) 1994, 1996-1997 Eric P. Allman.  All rights reserved.
  * Copyright (c) 1994
@@ -14,7 +14,7 @@
 #include <sendmail.h>
 #include <string.h>
 
-SM_RCSID("@(#)$Id: mime.c,v 8.130.2.1 2003/04/15 01:05:59 ca Exp $")
+SM_RCSID("@(#)$Id: mime.c,v 8.136 2004/03/22 18:21:34 ca Exp $")
 
 /*
 **  MIME support.
@@ -33,6 +33,11 @@ SM_RCSID("@(#)$Id: mime.c,v 8.130.2.1 2003/04/15 01:05:59 ca Exp $")
 **	the problem.
 */
 
+/* use "old" mime 7 to 8 algorithm by default */
+#ifndef MIME7TO8_OLD
+# define MIME7TO8_OLD	1
+#endif /* ! MIME7TO8_OLD */
+
 #if MIME8TO7
 static int	isboundary __P((char *, char **));
 static int	mimeboundary __P((char *, char **));
@@ -132,7 +137,7 @@ mime8to7(mci, header, e, boundaries, flags)
 	p = hvalue("Content-Transfer-Encoding", header);
 	if (p == NULL ||
 	    (pvp = prescan(p, '\0', pvpbuf, sizeof pvpbuf, NULL,
-			   MimeTokenTab)) == NULL ||
+			   MimeTokenTab, false)) == NULL ||
 	    pvp[0] == NULL)
 	{
 		cte = NULL;
@@ -154,7 +159,7 @@ mime8to7(mci, header, e, boundaries, flags)
 	}
 	if (p != NULL &&
 	    (pvp = prescan(p, '\0', pvpbuf, sizeof pvpbuf, NULL,
-			   MimeTokenTab)) != NULL &&
+			   MimeTokenTab, false)) != NULL &&
 	    pvp[0] != NULL)
 	{
 		if (tTd(43, 40))
@@ -763,11 +768,11 @@ mime_getchar(fp, boundaries, btp)
 			return SM_IO_EOF;
 		}
 
-		atbol = c == '\n';
-		if (c != SM_IO_EOF)
+		if (bp < &buf[sizeof buf - 2] && c != SM_IO_EOF)
 			*bp++ = c;
 	}
 
+	atbol = c == '\n';
 	buflen = bp - buf - 1;
 	if (buflen < 0)
 	{
@@ -985,7 +990,7 @@ mime7to8(mci, header, e)
 	p = hvalue("Content-Transfer-Encoding", header);
 	if (p == NULL ||
 	    (pvp = prescan(p, '\0', pvpbuf, sizeof pvpbuf, NULL,
-			   MimeTokenTab)) == NULL ||
+			   MimeTokenTab, false)) == NULL ||
 	    pvp[0] == NULL)
 	{
 		/* "can't happen" -- upper level should have caught this */
@@ -1064,54 +1069,41 @@ mime7to8(mci, header, e)
 			c1 = CHAR64(c1);
 			c2 = CHAR64(c2);
 
+#if MIME7TO8_OLD
+#define CHK_EOL if (*--fbufp != '\n' || (fbufp > fbuf && *--fbufp != '\r')) \
+			++fbufp;
+#else /* MIME7TO8_OLD */
+#define CHK_EOL if (*--fbufp != '\n' || (fbufp > fbuf && *--fbufp != '\r')) \
+		{					\
+			++fbufp;			\
+			pxflags |= PXLF_NOADDEOL;	\
+		}
+#endif /* MIME7TO8_OLD */
+
+#define PUTLINE64	\
+	do		\
+	{		\
+		if (*fbufp++ == '\n' || fbufp >= &fbuf[MAXLINE])	\
+		{							\
+			CHK_EOL;					\
+			putxline((char *) fbuf, fbufp - fbuf, mci, pxflags); \
+			pxflags &= ~PXLF_NOADDEOL;			\
+			fbufp = fbuf;					\
+		}	\
+	} while (0)
+
 			*fbufp = (c1 << 2) | ((c2 & 0x30) >> 4);
-			if (*fbufp++ == '\n' || fbufp >= &fbuf[MAXLINE])
-			{
-				if (*--fbufp != '\n' ||
-				    (fbufp > fbuf && *--fbufp != '\r'))
-				{
-					pxflags |= PXLF_NOADDEOL;
-					fbufp++;
-				}
-				putxline((char *) fbuf, fbufp - fbuf,
-					 mci, pxflags);
-				pxflags &= ~PXLF_NOADDEOL;
-				fbufp = fbuf;
-			}
+			PUTLINE64;
 			if (c3 == '=')
 				continue;
 			c3 = CHAR64(c3);
 			*fbufp = ((c2 & 0x0f) << 4) | ((c3 & 0x3c) >> 2);
-			if (*fbufp++ == '\n' || fbufp >= &fbuf[MAXLINE])
-			{
-				if (*--fbufp != '\n' ||
-				    (fbufp > fbuf && *--fbufp != '\r'))
-				{
-					pxflags |= PXLF_NOADDEOL;
-					fbufp++;
-				}
-				putxline((char *) fbuf, fbufp - fbuf,
-					 mci, pxflags);
-				pxflags &= ~PXLF_NOADDEOL;
-				fbufp = fbuf;
-			}
+			PUTLINE64;
 			if (c4 == '=')
 				continue;
 			c4 = CHAR64(c4);
 			*fbufp = ((c3 & 0x03) << 6) | c4;
-			if (*fbufp++ == '\n' || fbufp >= &fbuf[MAXLINE])
-			{
-				if (*--fbufp != '\n' ||
-				    (fbufp > fbuf && *--fbufp != '\r'))
-				{
-					pxflags |= PXLF_NOADDEOL;
-					fbufp++;
-				}
-				putxline((char *) fbuf, fbufp - fbuf,
-					 mci, pxflags);
-				pxflags &= ~PXLF_NOADDEOL;
-				fbufp = fbuf;
-			}
+			PUTLINE64;
 		}
 	}
 	else
